#include <stdio.h>
#include "y.tab.h"
#include "phones.h"

typedef enum lexState {
    EXPECT_NAME, EXPECT_PHONE, CHECK_MORE_PHONES, EXPECT_ADDRESS
} lexState;

lexState State = EXPECT_NAME;

int LineNo = 1;
char StringTable[200000], *StringTop = StringTable;
char *UniqueStrings[200], **UniqueTop = UniqueStrings;

#define TOP StringTop
char *START = StringTable;
#define CPUSH(c)							     \
{									     \
    *StringTop++ = c;							     \
    if (StringTop >= StringTable + sizeof(StringTable)) {		     \
	lexerror("string table overflow");				     \
    }									     \
}


char *storeunique(char *x)
{
    char **p = UniqueStrings;
    if (*x == '\0') return (char *) 0;
    while (p < UniqueTop) {
	if (! strcmp(x, *p) ) return *p;
	p++;
    }
    *p = x;
    UniqueTop++;
    if (UniqueTop >
	UniqueStrings + sizeof(UniqueStrings) / sizeof(UniqueStrings[0])) {
	lexerror("key table overflow");
	exit(1);
    }
    return x;
}

char *storestring(char until, char *what)
{
    int c;
    char *start = StringTop;

    while ((c = getc(stdin)) != until) {
	if (c == '\n') LineNo++;
	switch (c) {
	case EOF:
	    fprintf(stderr, "EOF found while scanning %s\n", what);
	    exit(1);
	default:
	    if (c == until) {
		*StringTop++ = '\0';
		return start;
	    } else {
		CPUSH(c);
	    }
	}
    }
    return start;
}

void unstorestring()
{
    while (*--StringTop);
}

yyerror()
{
    fprintf(stderr, "syntax error at line %d\n", LineNo);
    exit (1);
}

int get_key(char *string)
{
    
    char *y = storeunique(x);
    if (x != y) unstorestring();
    yylval.s = y;
    return DEFAULT_KEY;
}

int yylex()
{
    while (1) {
	int c = getc(stdin);

	switch (State) {

	case EXPECT_NAME:
	    /* Find if name or default key follows.
	     */
	    switch (c) {
	    case EOF:
		return 0;
	    case '\n':
		LineNo++;
		break;
	    case '@':
		return getKey("default key");
	    case '\t':
		lexerror("expected name; found tab");
	    case '(':
		lexerror("expected name; found comment");
	    default:
		ungetc(c, stdin);
		State = SCAN_NAME;
		break;
	    }
	    break;

	case SCAN_NAME:
	    /* Scan parts of name.
	     */
	    switch (c) {
	    case '\t':
		State = EXPECT_PHONE;
		break;
	    case ' ':
	    case '\n':
		lexerror("phone number must follow name");
	    default:
		
		yylval.s = storestring(" \t", "name");
		return NAME;
	    }
	    break;
	case EXPECT_PHONE:
	    switch (c) {
	    case ('\t'):
		break;
	    default:
		ungetc(c, stdin);
		yylval.s = storestring('\n', "phone number");
		State = CHECK_MORE_PHONES;
		return TELNO;
	    }
	    break;
	case CHECK_MORE_PHONES:
	    switch (c) {
	    case '\t':
		State = EXPECT_PHONE;
		break;
	    case '(':
		yylval.s = storestring(')', "comment");
		State = EXPECT_NAME;
		return COMMENT;
	    case '@': {
		char *x = storestring('\n', "default key");
		char *y = storeunique(x);
		if (x != y) unstorestring();
		yylval.s = y;
		return DEFAULT_KEY;
	    }
	    default:
		State = EXPECT_ADDRESS;
		ungetc(c, stdin);
		break;
	    }
	    break;
	case EXPECT_ADDRESS:
	    switch (c) {
	    case '\n':
		LineNo++;
		State = EXPECT_NAME;
		break;
	    case '(':
		yylval.s = storestring(')', "comment");
		State = EXPECT_NAME;
		return COMMENT;
	    case '@': {
		char *x = storestring('\n', "default key");
		char *y = storeunique(x);
		if (x != y) unstorestring();
		yylval.s = y;
		return KEY;
	    }
	    default:
		yylval.s = storestring('\n', "address");
		return ADDRESS;
	    }
	}
    }
}

