/*
    phoneformat - produce a compact TeX file for a simple address database.
    Copyright (C) 1989-1994  Luigi Semenzato

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Author's address:
    Computer Science Division
    University of California
    Berkeley, CA 94720
    luigi@cs.berkeley.edu

    This program was developed in the author's own time
    and is not connected in any way to the author's research
    at Berkeley.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "phones.h"

struct entry EntryTable[2000], *CE = EntryTable, *EndTable;
extern char *UniqueStrings[];
extern char **UniqueTop;
char *TokenString;
char *malloc();
int stringlistcompare();

int strcmp();

#define usage()	\
  fprintf(stderr, \
"usage: %s [-c] [-s] [-k] [-n <keyword>] [-y <keyword>]\n\
       [-R] [-4] [-P] [-f <size>] [-o]\n\
       [-Q <width> <height>] [-g] [-S <style>] [-U]\n", argv[0]), exit(1)


extern char TeXVarious[], *TeXSeparators[], TeXOversized[],
  *TeXFonts[], *TeXPFonts[], TeXThreecol[],
  TeXFourcol[], TeXQuad[], TeXGrid[], TeXEpilogue[];

extern int Nseparators, Nfonts, NPfonts;

int
  cflag = 0,	/* compact output */
  pflag = 0,	/* output pilot-addresses format */
  sflag = 0,	/* force sort on last name */
  Rflag = 0,	/* Rolodex format input */
  Pflag = 0,	/* Postscript fonts */
  Qflag = 0,	/* four quadrant output */
  fourFlag = 0,	/* four column output */
  gridFlag = 0, /* 3 x 4 grid output (credit-card size) */
  ultraFlag = 0,/* ultracompact: squeeze as much as possible */
  oflag = 0,	/* oversized letters in separators */
  kflag = 0,	/* print by key */
  key_yn = 0,	/* 1 = start with null keyword list, 2 = start with full */
  bogusflag;

int separator = 3;	/* alphabetical or keyword separator style */
int font = 0;		/* font (lower number = smaller) */

float DEFAULTWIDTH = 1.5;
float DEFAULTHEIGHT = 4.5;
float width, height;

#define XKEYS 50
char *xkey[XKEYS];
int xkeys = 0;

char *programName;

double atof();

int strcmpp(char **x, char **y)
{
    return strcmp(*x, *y);
}

main(ac, argv)
int ac;
char **argv;
{
        int argc;
	programName = argv[0];
	for (argc = 1; argc < ac; argc++) {
		if (argv[argc][0] == '-') {
			switch (argv[argc][1]) {
#define caseflag(char, flag) case char: flag = 1; break
			    caseflag('c', cflag);
			    caseflag('p', pflag);
			    caseflag('s', sflag);
			    caseflag('R', Rflag);
			    caseflag('P', Pflag);
			    caseflag('4', fourFlag);
			    caseflag('g', gridFlag);
			    caseflag('o', oflag);
			    caseflag('U', ultraFlag);
			    caseflag('k', kflag);
			case 'Q':
			    Qflag = 1;
			    argc++;
			    if (argc == ac) usage();
			    width = atof(argv[argc]);
			    if (width == 0) width = DEFAULTWIDTH;
			    argc++;
			    if (argc == ac) usage();
			    height = atof(argv[argc]);
			    if (height == 0) height = DEFAULTHEIGHT;
			    break;
			case 'S':
			    argc++;
			    if (argc == ac) usage();
			    separator = atoi(argv[argc]);
			    if (separator >= Nseparators) {
				fprintf(stderr,
					"%s: invalid separator style %d\n",
					argv[0], separator); exit(1);
			    }
			    break;
			case 'f':
			    argc++;
			    if (argc == ac) usage();
			    font = atoi(argv[argc]);
			    break;
			case 'y':
			    if (key_yn == 2) goto y;
			    key_yn = 1;
			    goto x;
			case 'n':
			    if (key_yn == 1) goto y;
			    key_yn = 2;
			x:
			    argc++;
			    if (argc == ac) usage();
			    xkey[xkeys++] = argv[argc];
			    if (xkeys == XKEYS) {
				fprintf(stderr, "%s: too many keys\n",
					argv[0]); exit(1);
			    }
			    break;
			y:
			    fprintf(stderr, "%s: both Y and N keys\n",
				    argv[0]); exit(1);
			default:
				fprintf(stderr, "%s: unknown flag -%c\n",
					argv[0], argv[argc][1]);
				usage();
			}
		} else {
			usage();
		}
	}
	lexinit();
	if (Rflag) {
		roloparse();
	} else {
		yyparse();
	}
	EndTable = CE;
	if (key_yn == 0) key_yn = 2;
	qsort(UniqueStrings, UniqueTop - UniqueStrings,
	      sizeof (char *),
	      (int (*)(const void *, const void *)) strcmpp);

	if ((!Rflag) || (sflag)) {
		qsort(EntryTable, EndTable-EntryTable,
		      sizeof(EntryTable[0]),
		      (int (*)(const void *, const void *)) stringlistcompare);
	}
#ifdef DEBUG
	for (CE=EntryTable; CE<EndTable; CE++) {
		printentry(CE);
	}
#endif
	if (pflag)
		pilotOutput();
	else
		format();
	exit(0);
}

#define MAXFILENAME 200
char rolofile[MAXFILENAME];
struct stringlist **dofield();
struct stringlist *fixentry();
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
struct stat statbuf;
char *memfile, *malloc();

char *savestring(string)
char *string;
{
    char *b = malloc(strlen(string) + 1);
    strcpy(b, string);
    return b;
}

/* roloparse -- parse assuming rolodex format */
roloparse()
{
    int rolosize;
    FILE *f, *fopen();
    char *currp;

    strncpy(rolofile, getenv("HOME"), MAXFILENAME);
    strncat(rolofile, "/.rolodex.dat", MAXFILENAME);
    stat(rolofile, &statbuf);
    rolosize = statbuf.st_size;
    rolosize += rolosize / 10;
    /* 10% for special characters */
    memfile = malloc(rolosize);
    currp = memfile;
    f = fopen(rolofile, "r");
    if (f == NULL) {
	fprintf(stderr, "cannot open file %s\n", rolofile);
	exit(1);
    }
    rolosize = read_and_quote(f, memfile, rolosize);

#define findeol(currp) {while (*currp++ != '\n'); currp[-1] = '\0';}

    for(;;) {
	struct stringlist **nexts, **nnexts, *centry;
	char buf[100];
	/* name */
	CE->sortname = CE->names = stringcons(currp, 0);
	findeol(currp);
	if (sflag) {
	    register char *last = currp-1;
	    while(*--last != ' ');
	    last++;
	    CE->sortname = stringcons(last, 0);
	}
	/* work phone */
	nexts = &CE->telnos;
	nnexts = dofield(&currp, nexts);
	if (nnexts != nexts) {
	    strcpy (buf, "w: ");
	    strcat (buf, (*nexts)->string);
	    (*nexts)->string = savestring(buf);
	}
	nexts = nnexts;
	/* home phone */
	nnexts = dofield(&currp, nexts);
	if (nnexts != nexts) {
	    strcpy (buf, "h: ");
	    strcat (buf, (*nexts)->string);
	    (*nexts)->string = savestring(buf);
	}
	nexts = nnexts;
	/* first address */
	nexts = &CE->address;
	nexts = dofield(&currp, nexts);
	/* second address */
	nexts = dofield(&currp, nexts);
	/* third address */
	nexts = dofield(&currp, nexts);
	/* electronic address */
	nexts = dofield(&currp, nexts);
	/* is this the end of the record? */
	if (*currp++ == '\n') goto next;
	/* now find two consecutive newlines */
	for (;;) {
	    if (*currp++ == '\n')
	      if (*currp++ == '\n')
		break;
	}
    next:
	CE++;
	if (currp >= memfile+rolosize) break;
    }
}

int
read_and_quote(f, buffer, size)
register FILE *f;
register char *buffer;
int size;
{
    register int c;
    char *start = buffer;
    
    while ((c = getc(f)) != EOF) {
	switch (c) {
	case '%':
	case '$':
	case '#':
	case '&':
	case '_':
	case '^':
	    *buffer++ = '\\';
	    /* fall into ... */
	default:
	    *buffer++ = c;
	}
    }
    if (buffer > start + size) {
	fprintf(stderr, "too many special characters -- buffer overflow\n");
	exit(1);
    }
    return buffer - start;
}

struct stringlist **
dofield(currp, nexts)
char **currp;
struct stringlist **nexts;
{
	struct stringlist *centry;
	register char *p = *currp;

	if (*p != '\n') {
		*nexts = stringcons(p, 0);
		centry = *nexts;
		nexts = &(*nexts)->next;
	} else centry = NULLS;
	findeol(p);
	if ((centry = fixentry(centry)) != NULLS) {
		nexts = &(centry->next);
	}
	*currp = p;
	return nexts;
}

struct stringlist *
fixentry(slp)
struct stringlist *slp;
{
	char *p;
	if (slp == NULLS) return slp;
	p = (slp)->string;
	while (*p) {
		if (*p == ';') {
			*p = '\0';
			p++;
			(slp)->next = stringcons(p, 0);
			slp = slp->next;
		} else {
			p++;
		}
	}
	return slp;
}

int findKey(k)
char *k;
{
    struct stringlist *slp = CE->keys;
    for (; slp; slp = slp->next) {
	if (! strcmp(slp->string, k)) return 1;
    }
    return 0;
}

int Section = 0;

formatEntry(key)
char *key;
{
    struct stringlist *more;

    if (key) {
	int found = 0;
	char **p = xkey;
	for (; p < xkey + xkeys; p++) {
	    if (! strcmp(*p, key)) { found = 1; break; }
	}
	if (key_yn == 1 && ! found) return;
	if (key_yn == 2 && found) return;
	if (Section != (int) key) {
	    if (Section != 0) {
		if (ultraFlag) {
		} else if (cflag) {
		    printf("\\vskip .3em plus .06em\n");
		} else {
		    printf("\\vskip .6em plus .12em\n");
		    printf("\\par\\penalty-200\n");
		}
	    }
	    Section = (int) key;
	    printf("\\separator{%s}\n", key);
	    
	}
    } else {
	if (Section != CE->sortname->string[0]) {
	    if (Section != 0) {
		if (ultraFlag) {
		} else if (cflag) {
		    printf("\\vskip .3em plus .06em\n");
		} else {
		    printf("\\vskip .6em plus .12em\n");
		    printf("\\par\\penalty-200\n");
		}
	    }
	    Section = CE->sortname->string[0];
	    printf("\\separator{%c}\n", Section);
	}
    }

    if (xkeys) {
	int found = 0;
	char **p = xkey;
	for (; p < xkey + xkeys; p++) {
	    if (findKey(*p)) { found = 1; break; }
	}
	if (key_yn == 1 && ! found) return;
	if (key_yn == 2 && found) return;
    }

    if (kflag && key && ! findKey(key)) return;
    
    if (!ultraFlag && !cflag) {
	printf("\\smallskip");
    }
    printf("\\firstline{\\bf ");
    more = CE->names;
    do {
	if (more->next != NULLS) {
	    printf("%s ", more->string);
	} else {
	    printf("{%s}", more->string);
	}
    } while ((more = more->next) != NULLS);
    printf("}{");
    printf("%s}\n",
	   CE->telnos? CE->telnos->string : "");
    if (CE->telnos) {
	more = CE->telnos->next;
    } else {
	more = NULLS;
    }
    while (more != NULLS) {
	printf("\\rightline{%s}\n", more->string);
	more = more->next;
    }
    if (cflag) return;
    more = CE->address;
    if (ultraFlag && more != NULLS) {
	printf("\\noindent\n");
    }
    while (more != NULLS) {
	if (!ultraFlag) {
	    printf("\\noindent %s\\hfil\\break\n", more->string);
	} else {
	    printf("%s\n", more->string);
	    if (more->next != NULLS) {
		printf("--- ");
	    } else {
		printf("\\hfil\\break\n");
	    }
	}
	more = more->next;
    }
    more = CE->comment;
    if (more == NULLS) return;
    printf("\\noindent{\\it ");
    while (more != NULLS) {
	printf("%s\n", more->string);
	more = more->next;
    }
    printf("}\n");
    if (ultraFlag) printf("\\hfil\\break\n");
}

format()
{
	char **fonts;
	int nfont;

	printf("%s", TeXVarious);
	printf("%s", TeXSeparators[separator]);
	if (Pflag) {
	    fonts = TeXPFonts;
	    nfont = NPfonts;
	} else {
	    fonts = TeXFonts;
	    nfont = Nfonts;
	}
	if (font >= nfont) {
	    fprintf(stderr,
		    "%s: invalid font choice %d\n",
		    programName, font); exit(1);
	}
	printf("%s", fonts[font]);
	if (oflag) {
	    printf("%s", TeXOversized);
	}
	if (Qflag) {
	    float fullhsize = 2.1 * width;
	    float leftmargin = (8.5 - 2 * fullhsize) / 4;
	    float FULLHSIZE = 2 * fullhsize + 2 * leftmargin;
	    float hoffset = leftmargin - 1;
	    float fullvsize = height;	
	    float topmargin = (11 - 2 * fullvsize) / 4;
	    float FULLVSIZE = 2 * fullvsize + 2 * topmargin;
	    float voffset = topmargin - 1.1;	/* I wonder why */
	    printf("%s", TeXQuad);
	    printf("\\voffset=%5.2fin\n", voffset);
	    printf("\\hoffset=%5.2fin\n", hoffset);
	    printf("\\FULLHSIZE=%5.2fin\n", FULLHSIZE);
	    printf("\\FULLVSIZE=%5.2fin\n", FULLVSIZE);
	    printf("\\fullhsize=%5.2fin\n", fullhsize);
	    printf("\\fullvsize=%5.2fin\n", fullvsize);
	    printf("\\hsize=%5.2fin\n", width);
	    printf("\\vsize=%5.2fin\n", height);
	} else if (fourFlag) {
	    printf("%s", TeXFourcol);
	} else if (gridFlag) {
	    printf("%s", TeXGrid);
	} else {
	    printf("%s", TeXThreecol);
	}
	if (kflag) {
	    char **p = UniqueStrings;
	    for (; p < UniqueTop; p++) {
		for (CE=EntryTable; CE<EndTable; CE++) {
		    formatEntry(*p);
		}
	    }
	} else {
	    for (CE=EntryTable; CE<EndTable; CE++) {
		formatEntry(0);
	    }
	}
	printf("%s", TeXEpilogue);
}

#define prefix(s, tag) (strncmp((s)->string, tag, sizeof tag - 1) == 0)

#define pn(s, l, n) do { \
	char *p; \
	if (n) \
		fprintf(stderr, "Too many "#n": %s\n", (s)->string); \
	for (p = (s)->string + (l); isspace(*p); p++); \
	(n) = p; \
} while (0)

#define pnother(s, l) do { \
	if (!other) { \
		pn(s, l, other); \
	} else { \
		pncustom(s, l); \
	} \
} while (0)

#define pncustom(s, l) do { \
	if (!custom1) { \
		pn(s, l, custom1); \
	} else if (!custom2) { \
		pn(s, l, custom2); \
	} else if (!custom3) { \
		pn(s, l, custom3); \
	} else if (!custom4) { \
		pn(s, l, custom4); \
	} else { \
		fprintf(stderr, "Too many other: %s\n", (s)->string); \
	} \
} while (0)

pilotOutput()
{
	char *p;
	struct stringlist *s;
	int first;

	for (CE=EntryTable; CE<EndTable; CE++) {
		char *company = 0;
		char *work = 0;
		char *home = 0;
		char *fax = 0;
		char *other = 0;
		char *email = 0;
		char *address = 0;
		char *city = 0;
		char *state = 0;
		char *zip = 0;
		char *country = 0;
		char *custom1 = 0;
		char *custom2 = 0;
		char *custom3 = 0;
		char *custom4 = 0;
		char *note = 0;
		printf("\"%s\"", CE->sortname->string);
		printf(",\"");
		first = 1;
		for (s = CE->names; s != NULL; s = s->next) {
			if (strcmp(s->string, CE->sortname->string) == 0)
				continue;
			if (first) {
				first = 0;
			} else {
				printf(" ");
			}
			printf("%s", s->string);
		}
		printf("\",\"\"");
		for (s = CE->telnos; s != NULL; s = s->next) {
			if (prefix(s, "work")) {
				pn(s, 4, work);
				continue;
			}
			if (prefix(s, "main")) {
				pn(s, 4, work);
				continue;
			}
			if (prefix(s, "home")) {
				pn(s, 4, home);
				continue;
			}
			if (prefix(s, "fax")) {
				pn(s, 3, fax);
				continue;
			}
			if (prefix(s, "other")) {
				pnother(s, 5);
				continue;
			}
			if (prefix(s, "pager")) {
				pnother(s, 0);
				continue;
			}
			if (prefix(s, "cellular")) {
				pnother(s, 0);
				continue;
			}
			if (prefix(s, "secretary")) {
				pnother(s, 0);
				continue;
			}
			if (isdigit(s->string[0])) {
				if (!home)
					pn(s, 0, home);
				else
					pnother(s, 0);
				continue;
			}
			fprintf(stderr, "Can't classify number: %s\n",
				s->string);
		}
		for (s = CE->address; s != NULL; s = s->next) {
			char *p;
			if (strcmp(s->string, "{}") == 0) {
				continue;
			}
			if (prefix(s, "{\\tt")) {
				for (p = s->string + strlen(s->string) - 1;
				     p >= s->string && *p == '}';
				     p--);
				p[1] = 0;
				pn(s, 4, email);
				continue;
			}
			if (!address && !company &&
			    !isdigit(s->string[0]) &&
			    !strstr(s->string, " Box")) {
				pn(s, 0, company);
				continue;
			}
			if (!address) {
				pn(s, 0, address);
				continue;
			}
			if (address && !city && (p = strchr(s->string, ','))) {
				*p = 0;
				pn(s, 0, city);
				for (p++; isspace(*p); p++);
				s->string = p;
				for (p++; *p && !isspace(*p); p++);
				*p = 0;
				pn(s, 0, state);
				for (p++; isspace(*p); p++);
				s->string = p;
				for (p++; *p && !isspace(*p); p++);
				*p = 0;
				pn(s, 0, zip);
				continue;
			}
			pncustom(s, 0);
		}
		for (s = CE->comment; s != NULL; s = s->next) {
			pncustom(s, 0);
		}
		printf(",\"%s\"", company ? company : "");
		printf(",\"%s\"", work ? work : "");
		printf(",\"%s\"", home ? home : "");
		printf(",\"%s\"", fax ? fax : "");
		printf(",\"%s\"", other ? other : "");
		printf(",\"%s\"", email ? email : "");
		printf(",\"%s\"", address ? address : "");
		printf(",\"%s\"", city ? city : "");
		printf(",\"%s\"", state ? state : "");
		printf(",\"%s\"", zip ? zip : "");
		printf(",\"%s\"", country ? country : "");
		printf(",\"%s\"", custom1 ? custom1 : "");
		printf(",\"%s\"", custom2 ? custom2 : "");
		printf(",\"%s\"", custom3 ? custom3 : "");
		printf(",\"%s\"", custom4 ? custom4 : "");
		/*
		printf(",\"");
		first = 1;
		for (s = CE->comment; s != NULL; s = s->next) {
			if (first) {
				first = 0;
			} else {
				printf(" ");
			}
			printf("%s", s->string);
		}
		printf("\"");
		*/
		printf(",\"\"");
		printf(",\"0\"\n");
	}
}

struct stringlist *
stringcons(strp, next)
char *strp;
struct stringlist *next;
{
	struct stringlist *temp =
	  (struct stringlist *) malloc(sizeof(struct stringlist));
	temp->next = next;
	temp->string = strp;
	return temp;
}

stringlistcompare(a, b)
struct entry *a, *b;
{
	int t = strcmp(a->sortname->string, b->sortname->string);
	if (t != 0) return t;
	else return strcmp(a->names->string, b->names->string);
}

printentry(entry)
struct entry *entry;
{
	printf("\n");
	printf("SORTNAME = \"%s\"\n", entry->sortname->string);
	printf("\n");
	printf("NAMES = ");
	printlist(entry->names);
	printf("\n");
	printf("TELNOS = ");
	printlist(entry->telnos);
	printf("\n");
	printf("ADDRESS = ");
	printlist(entry->address);
	printf("\n");
	printf("COMMENT = ");
	printlist(entry->comment);
	printf("\n");
	printf("\n");
}
	

printlist(strl)
struct stringlist *strl;
{
	while(strl) {
		printf("%s ", strl->string);
		strl = strl->next;
	}
}
